/* pile up the image files.
For int8, int16, float32 bit & mono, RGB, RGBG color images.

2004/06/28 by NAKAMURA Kazushi
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include "pile.h"

enum{
	MEAN,
	MEDIUM
} alg;

enum{
	PRW,
	PPM,
	PGM
} outfmt;

#define SZ	256
#define THRETH	50000

/*** defaults ***/
int sx=256,sy=256;	//ֹ碌׻ϰ
int px=16,py=16;	//ֹ碌ưϰ
int ax= -1,ay= -1;	//ֹ碌濴
int verb=0;		//verbose mode
int odepth=16;		//ϳĴ 8,16,-32(float)

/*** global ***/
int ix,iy;		// Input image size
int ox,oy;		// Output image size
int noi;		//ϥե
int basis;		//ֹ碌ե#
int *dx,*dy;		//
int maxdx,mindx;
int maxdy,mindy;
INT16 *I,*O;
INT32 *W;
FILE **fp,*ofp;

void usage(){
	fprintf(stderr,"Usage: pile [-v -a <alg> -s <size> -p <pivot> "
		"-x <x> -y <y> -d <depth> -t <threshold>] -o out.fmt files...\n"
		"\t<alg>={mean,medium,dif}\n"
		"\t<depth>={8,16,-32}\n"
		"\t.fmt={prw,ppm,pgm}\n");
	exit(1);
}

int main(int ac, char **av){
int ch,d,colors;
int x,y,x0,y0;
unsigned int d0;
char *ofname=NULL;
char dmy[SZ];
int i,j,c;
int xx,yy;
char r;
int th=THRETH;

/*** defaults ***/
	alg=MEAN;
	outfmt=PRW;
	colors=3;

/*** option ***/
	while((ch=getopt(ac,av,"va:t:s:p:x:y:o:")) != -1)
		switch(ch){
		case 'v':
			verb++;
			break;
		case 'a':
			if(!strcasecmp(optarg,"mean")){
				alg=MEAN;
			}else if(!strcasecmp(optarg,"medium")){
				alg=MEDIUM;
			}else usage();
			break;
		case 's':
			sx=sy=atoi(optarg);
			break;
		case 'p':
			px=py=atoi(optarg);
			break;
		case 'x':
			ax=atoi(optarg);
			break;
		case 'y':
			ay=atoi(optarg);
			break;
		case 'd':
			odepth=atoi(optarg);
			break;
		case 't':
			th=atoi(optarg);
			break;
		case 'o':
			ofname=optarg;
			if(ofname[strlen(ofname)-4]!='.'){
				fprintf(stderr,"Invalid surfix!\n");
				exit(1);
			}
			ofp=fopen(ofname,"w");
			if(ofp==NULL){
				perror(ofname); exit(1);
			}
			switch(ofname[strlen(ofname)-2]){
			case 'r':
				outfmt=PRW;
				break;
			case 'p':
				outfmt=PPM;
				break;
			case 'g':
				outfmt=PGM;
				break;
			default:
				fprintf(stderr,"Invalid surfix!\n");
				exit(1);
			}
			break;
		default:
			usage();
		}
	ac-=optind;
	av+=optind;
	if(ofname==NULL) usage();

	noi=ac; if(noi<1) usage();
	basis=noi/2;

	if(verb) printf("v=%d alg=%d sx=%d px=%d ax=%d ay=%d th=%d fmt=%d %s ac=%d noi=%d base=%d\n"
		,verb,alg,sx,px,ax,ay,th,outfmt,ofname,ac,noi,basis);

	dx=malloc(noi*sizeof(int));
	dy=malloc(noi*sizeof(int));
	fp=malloc(noi*sizeof(FILE **));

	for(i=0; i<noi; i++){
		fp[i]=fopen(av[i],"r");
		if(fp[i]==NULL){
			perror(av[i]); exit(1);
		}
	}
	readi(fp[basis],0,&ix,&iy,th);

	////align
	for(i=0; i<noi; i++){
		if(i==basis){
        		printf("%2d %s %+3d %+3d %15u *\n",i,av[i],0,0,0);
			dx[i]=dy[i]=0; continue;
		}
		readi(fp[i],1,&x,&y,th);
		if(x!=ix || y!=iy){
			fprintf(stderr,"Error: size missmatch. %s\n",av[i]);
			exit(1);
		}
		d0=(-1);	// MAX
		d=pivot(&x0,&y0,d0);
		dx[i]=x0; dy[i]=y0;
		r=(abs(x0)==px || abs(y0)==py)? '!':' ';
        	printf("%2d %s %+3d %+3d %15u %c\n",i,av[i],x0,y0,d,r);
		if(x0>maxdx){
			maxdx=x0;
		}else if(x0<mindx){
			mindx=x0;
		}
		if(y0>maxdy){
			maxdy=y0;
		}else if(y0<mindy){
			mindy=y0;
		}
	}
#ifdef DEBUG
	for(int i=0; i<noi; i++){
		printf("%2d %+3d %+3d\n",i,dx[i],dy[i]);
	}
#endif

	////output size
	ox=ix-(maxdx-mindx);
	oy=iy-(maxdy-mindy);

	//DEBUG only PPM
	fprintf(ofp,"P6\n%d %d\n%d\n", ox,oy, (1<<odepth)-1);
	fflush(ofp);

        I=malloc(ix * iy * sizeof(*I) * colors);
        W=malloc(ox * oy * sizeof(*W) * colors);
        O=malloc(ox * oy * sizeof(*O) * colors);

	////pile
	for(i=0; i<noi; i++){
//		fseek(fp[i],0L,SEEK_SET);
		rewind(fp[i]);
//		fgets(dmy,SZ,fp[i]); fgets(dmy,SZ,fp[i]); fgets(dmy,SZ,fp[i]);
		memset(dmy,0,SZ);
		fgets(dmy,SZ,fp[i]);
		memset(dmy,0,SZ);
		fgets(dmy,SZ,fp[i]);
		memset(dmy,0,SZ);
		fgets(dmy,SZ,fp[i]);
#if 1
		for(j=0; j<dy[i]-mindy; j++)
			fseek(fp[i],ix*sizeof(*I)*colors,SEEK_CUR);
#endif
	}

	for(yy=0;yy<oy;yy++){
		memset(W,0,ox*sizeof(*W)*colors);
		for(i=0; i<noi; i++){
			fread(I,ix*sizeof(*I)*colors,1,fp[i]);
			swab(I,I,ix*sizeof(*I)*colors);
			for(xx=0;xx<ox;xx++){
				for(c=0;c<colors;c++){
					W[xx*colors+c]+=I[(xx+dx[i]-mindx)*colors+c];
				}
			}
		}
		//DEBUG only ûʿ
		for(xx=0;xx<ox;xx++){
			for(c=0;c<colors;c++){
				O[xx*colors+c]=W[xx*colors+c]/noi;
			}
		}
		swab(O,O,ox*sizeof(*O)*colors);
		fwrite(O,ox*sizeof(*O)*colors,1,ofp);
		fflush(ofp);
	}

	return 0;
}
